// optimize.cpp

#include <iostream>
#include <algorithm>
#include "optimize.h"

using namespace boost::numeric;

double optimize_epsilon;

box::box(const box& b)
{
	interval[0] = b.interval[0];
	interval[1] = b.interval[1];
	interval[2] = b.interval[2];
	value = b.value;
}

box::box(INTERVAL *it, INTERVAL v)
{
	interval[0] = it[0];
	interval[1] = it[1];
	interval[2] = it[2];
	value = v;
}

box& box::operator=(const box& b)
{
	interval[0] = b.interval[0];
	interval[1] = b.interval[1];
	interval[2] = b.interval[2];
	value = b.value;
	
	return *this;
}

double box::maxwidth(void)
{
	return std::max(std::max(width(interval[0]), width(interval[1])), width(interval[2]));
}

int minimum(INTERVAL &ret, std::multiset<box> &lb, p_func f, int n_ineq, p_func *p, int n_eq, p_func *q, p_func_c *qc)
{
	int i, j, k, flag, flag2, count;
	INTERVAL it, s[3][2];
	double w, u_bound;
	box b;
	std::multiset<box> ls2;

	u_bound = ret.upper();

	count = 0;
	while(!lb.empty()) {
		count++;
		b = *(lb.begin());
		lb.erase(lb.begin());

		// if b has no feasible points, then delete b
		flag = 0;
		flag2 = 1;
		for(i = 0; i < n_ineq; i++) {
			p[i](it, b.interval);
			if(it.lower() > 0) {
				flag = 1;
				break;
			}
			if(flag2 && it.upper() > 0) {
				flag2 = 0;
			}
		}
		if(flag)
			continue;
		flag = 0;
		for(i = 0; i < n_eq; i++) {
			q[i](it, b.interval);
			if(!zero_in(it)) {
				flag = 1;
				break;
			}
		}
		if(flag) {
			continue;
		}
		// if b has a feasible point, then update u_bound
		if(qc != NULL) {
			if(flag2) {
				for(i = 0; i < n_eq && flag2; i++) {
					flag2 = qc[i](b.interval);
				}
			}
			if(flag2) {
				if(b.value.upper() < u_bound) {
					u_bound = b.value.upper();
				}		
			}
		}
			
		if(b.maxwidth() < optimize_epsilon) {
			ls2.insert(b);
			continue;
		}
		for(i = 0; i < 3; i++) {
			w = median(b.interval[i]);
			s[i][0] = INTERVAL(b.interval[i].lower(), w);
			s[i][1] = INTERVAL(w, b.interval[i].upper());
		}
		for(i = 0; i < 2; i++) {
			for(j = 0; j < 2; j++) {
				for(k = 0; k < 2; k++) {
					b.interval[0] = s[0][i];
					b.interval[1] = s[1][j];
					b.interval[2] = s[2][k];
					f(b.value, b.interval);
					if(b.value.lower() <= u_bound) {
						lb.insert(b);
					}
				}
			}
		}
	}
	
//	std::cout << "count:" << count << std::endl;
	
	if(ls2.empty()) {
		lb.swap(ls2);
		return 0;
	}

	std::multiset<box>::iterator itr = ls2.begin();
	it = itr->value;
	for(itr++; itr != ls2.end(); itr++) {
		it = hull(it, itr->value);
	}
	ret.assign(ret.lower(), u_bound);
	ret = intersect(ret, it);
	
	lb.swap(ls2);

	return 1;
}

void pol_gradient(INTERVAL *ret, int deg, double *f, INTERVAL *x)
{
	int i, j, k;
	INTERVAL t1, t2;
	
	ret[0] = 0.0;
	for(i = deg; i > 0; i--) {
		t1 = 0.0;
		for(j = deg; j >= 0; j--) {
			t2 = 0.0;
			for(k = deg; k >= 0; k--) {
				t2 *= x[2];
				t2 += f[deg_conv(deg, i, j, k)] * i;
			}
			t1 *= x[1];
			t1 += t2;
		}
		ret[0] *= x[0];
		ret[0] += t1;
	}

	ret[1] = 0.0;
	for(i = deg; i >= 0; i--) {
		t1 = 0.0;
		for(j = deg; j > 0; j--) {
			t2 = 0.0;
			for(k = deg; k >= 0; k--) {
				t2 *= x[2];
				t2 += f[deg_conv(deg, i, j, k)] * j;
			}
			t1 *= x[1];
			t1 += t2;
		}
		ret[1] *= x[0];
		ret[1] += t1;
	}
	
	ret[2] = 0.0;
	for(i = deg; i >= 0; i--) {
		t1 = 0.0;
		for(j = deg; j >= 0; j--) {
			t2 = 0.0;
			for(k = deg; k > 0; k--) {
				t2 *= x[2];
				t2 += f[deg_conv(deg, i, j, k)] * k;
			}
			t1 *= x[1];
			t1 += t2;
		}
		ret[2] *= x[0];
		ret[2] += t1;
	}

	return;
}

void pol_value_Horner(INTERVAL &ret, int deg, double *f, INTERVAL *x)
{
	int i, j, k;
	INTERVAL t1, t2;
	
	ret = 0.0;
	for(i = deg; i >= 0; i--) {
		t1 = 0.0;
		for(j = deg; j >= 0; j--) {
			t2 = 0.0;
			for(k = deg; k >= 0; k--) {
				t2 *= x[2];
				t2 += f[deg_conv(deg, i, j, k)];
			}
			t1 *= x[1];
			t1 += t2;
		}
		ret *= x[0];
		ret += t1;
	}

	return;
}

void pol_value(INTERVAL &ret, int deg, double *f, INTERVAL *x)
{
	int i;
	double xc[3];
	INTERVAL y[3];
	INTERVAL grad[3];
	
	for(i = 0; i < 3; i++) {
		xc[i] = median(x[i]);
		y[i] = xc[i];
	}
	pol_value_Horner(ret, deg, f, y);
	pol_gradient(grad, deg, f, x);
	for(i = 0; i < 3; i++) {
		ret += grad[i] * (x[i] - xc[i]);
	}
	
	pol_value_Horner(y[0], deg, f, x);
	ret = intersect(ret, y[0]);
		
	return;
}

int pol_subst(INTERVAL *ret, int deg, double *f, INTERVAL *x)
{
	int i, j, k, ret_deg = deg;
	bool flag = false;
	INTERVAL t1, t2;
	
	for(k = deg; k >= 0; k--) {
		t1 = 0.0;
		for(i = deg; i >= 0; i--) {
			t2 = 0.0;
			for(j = deg; j >= 0; j--) {
				t2 *= x[1];
				t2 += f[deg_conv(deg, i, j, k)];
			}
			t1 *= x[0];
			t1 += t2;
		}
		if(flag || t1 != 0.0) {
			flag = true;
		} else {
			ret_deg--;
		}
		ret[k] = t1;
	}

	return ret_deg;
}

int pol_sturm(int deg, INTERVAL *f, INTERVAL x)
{
	int i, j, ret;
	INTERVAL y[3];
	double xc[2];
	
	for(i = 0; i < 2; i++) {
		y[i] = 0.0;
	}
	
	xc[0] = x.lower();
	xc[1] = x.upper();
	
	for(i = 0; i < 2; i++) {
		for(j = deg; j >= 0; j--) {
			y[i] *= xc[i];
			y[i] += f[j];
		}
	}
	if(y[0].lower() > 0 && y[1].upper() < 0) {
		ret = 1;
	} else if(y[0].upper() < 0 && y[1].lower() > 0) {
		ret = 1;
	} else {
		ret = -1;
	}
	
	return ret;
}
