// optimize.h

#ifndef OPTIMIZE_H
#define OPTIMIZE_H

#include <boost/numeric/interval.hpp>
#include <set>

typedef boost::numeric::interval<double,
	boost::numeric::interval_lib::policies<boost::numeric::interval_lib::save_state_nothing<
	boost::numeric::interval_lib::rounded_transc_std<double> >,
	boost::numeric::interval_lib::checking_strict<double> > > INTERVAL;

typedef void (*p_func)(INTERVAL&, INTERVAL*);
typedef int (*p_func_c)(INTERVAL*);

extern double optimize_epsilon;

class box {
public:
	box() {}
	box(const box& b);
	box(INTERVAL *it, INTERVAL v);
	~box() {};
	
	box& operator=(const box& b);
	bool operator<(const box& b) const {
		return (value.lower() < b.value.lower()) ||
			((value.lower() == b.value.lower()) && (value.upper() < b.value.upper()));
	}
	
	double maxwidth(void);
	
	INTERVAL interval[3];
	INTERVAL value;
};

// Set the value of ret to the minimum of the function f subject to
// p_i(x) <= 0, q_i(x) = 0. The range of f is lb.
// Set the value of lb to the intervals where the minimum attains.
// If ret is set, we assume that the minimum of f is contained in ret.
// Return 0 if it is proved that there does not exist a point satisfying
// the constraints, and 1 otherwise.
int minimum(INTERVAL &ret, std::multiset<box> &lb, p_func f, int n_ineq, p_func *p, int n_eq, p_func *q, p_func_c *qc);

// Initialize workspace for routines for polynomials.
void pol_initialize(void);

// Set the value of ret to the value of the polynomial of degree <= deg
// defined by the array f at the interval x.
void pol_value(INTERVAL &ret, int deg, double *f, INTERVAL *x);

// Substitute x1 and x2 in f, and set ret to it.
// deg is the degree of f.
// Return the degree of ret with respect to x3.
int pol_subst(INTERVAL *ret, int deg, double *f, INTERVAL *x);

// Return 1 if we can verify that f has a root in the interval x.
// If we cannot verify it, return -1;
int pol_sturm(int deg, INTERVAL *f, INTERVAL x);

inline int deg_conv(int d, int i, int j, int k)
{
	return (i * (d + 1) + j) * (d + 1) + k;
}
#endif
